/*	$NetBSD: ugensa.c,v 1.13 2007/12/11 12:36:02 lukem Exp $	*/
/*	$FreeBSD$	*/

/*
 * Copyright (c) 2004, 2005 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Roland C. Dowdeswell <elric@netbsd.org>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/tty.h>
#include <sys/taskqueue.h>

#include <dev/usb/usb_port.h>
#include <dev/usb/usb.h>

#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>

#include "usbdevs.h"

#include "ugensa.h"

#include <dev/usb/ucomvar.h>

/* XXXrcd: heh */
#define UGENSA_DEBUG 1

#ifdef UGENSA_DEBUG
#define DPRINTF(x)	if (ugensadebug) printf x
#define DPRINTFN(n,x)	if (ugensadebug>(n)) printf x
int ugensadebug = 1;
#else
#define DPRINTF(x)
#define DPRINTFN(n,x)
#endif

#define UGENSA_MODVER		1	/* module version */

#define UGENSA_CONFIG_INDEX	0
#define UGENSA_IFACE_INDEX	0
#define UGENSA_IBUFSIZE		1024
#define UGENSA_OBUFSIZE		1024

struct ugensa_softc {
	struct ucom_softc	sc_ucom;

	struct task		sc_task;
};

static void	ugensa_notify(void *, int);

struct ucom_callback ugensa_callback = {
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

static const struct usb_devno ugensa_devs[] = {
	{ USB_VENDOR_AIRPRIME, USB_PRODUCT_AIRPRIME_PC5220 },
	{ USB_VENDOR_DELL, USB_PRODUCT_DELL_HSDPA },
	{ USB_VENDOR_KYOCERA2, USB_PRODUCT_KYOCERA2_CDMA_MSM_K },
	{ USB_VENDOR_NOVATEL, USB_PRODUCT_NOVATEL_CDMA_MODEM },
	{ USB_VENDOR_NOVATEL2, USB_PRODUCT_NOVATEL2_FLEXPACKGPS },
	{ USB_VENDOR_SIERRA, USB_PRODUCT_SIERRA_AIRCARD580 },
	{ USB_VENDOR_YISO, USB_PRODUCT_YISO_C893 },
	{ USB_VENDOR_ZTE, USB_PRODUCT_ZTE_AC8700 },
};
#define ugensa_lookup(v, p) usb_lookup(ugensa_devs, v, p)

static device_probe_t ugensa_match;
static device_attach_t ugensa_attach;
static device_detach_t ugensa_detach;

static device_method_t ugensa_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe, ugensa_match),
	DEVMETHOD(device_attach, ugensa_attach),
	DEVMETHOD(device_detach, ugensa_detach),
	{ 0, 0 }
};

static driver_t ugensa_driver = {
	"ugensa",
	ugensa_methods,
	sizeof (struct ugensa_softc)
};

DRIVER_MODULE(ugensa, uhub, ugensa_driver, ucom_devclass, usbd_driver_load, 0);
MODULE_DEPEND(ugensa, usb, 1, 1, 1);
MODULE_DEPEND(ugensa, ucom, UCOM_MINVER, UCOM_PREFVER, UCOM_MAXVER);
MODULE_VERSION(ugensa, UGENSA_MODVER);

static int
ugensa_match(device_t self)
{
	struct usb_attach_arg *uaa = device_get_ivars(self);

	return (ugensa_lookup(uaa->vendor, uaa->product) != NULL ?
		UMATCH_VENDOR_PRODUCT : UMATCH_NONE);
}

static int
ugensa_attach(device_t self)
{
	usbd_interface_handle iface;
	struct ugensa_softc *sc = device_get_softc(self);
	struct usb_attach_arg *uaa = device_get_ivars(self);
	usbd_device_handle dev;
	struct ucom_softc *ucom;
	usb_interface_descriptor_t *id;
	usb_endpoint_descriptor_t *ed;
	const char *devname;
	usbd_status err;
	int i;

        if ((sc == NULL) || (uaa == NULL)) {
            return ENOMEM;
        }

	dev = uaa->device;

	ucom = &sc->sc_ucom;
	ucom->sc_dev = self;
	ucom->sc_udev = dev;

	devname = device_get_nameunit(self);

	DPRINTF(("ugensa_attach: sc = %p\n", sc));

	/* Move the device into the configured state. */
	err = usbd_set_config_index(dev, UGENSA_CONFIG_INDEX, 1);
	if (err) {
		device_printf(self, "failed to set configuration, err=%s\n",
			      usbd_errstr(err));
		goto error;
	}

	/* get the (first/common) interface */
	err = usbd_device2interface_handle(dev, UGENSA_IFACE_INDEX, &iface);
	if (err) {
		printf("\n%s: failed to get interface, err=%s\n",
		       devname, usbd_errstr(err));
		goto error;
	}

	id = usbd_get_interface_descriptor(iface);

	ucom->sc_parent = sc;
	ucom->sc_portno = UCOM_UNK_PORTNO;
	/* bulkin, bulkout set above */
	ucom->sc_ibufsize = UGENSA_IBUFSIZE;
	ucom->sc_obufsize = UGENSA_OBUFSIZE;
	ucom->sc_ibufsizepad = UGENSA_IBUFSIZE;
	ucom->sc_iface = iface;
	ucom->sc_opkthdrlen = 0;
	ucom->sc_callback = &ugensa_callback;

	/* Find the bulk{in,out} endpoints */
	ucom->sc_bulkin_no = ucom->sc_bulkout_no = -1;
	for (i = 0; i < id->bNumEndpoints; i++) {
		int addr, dir, attr;

		ed = usbd_interface2endpoint_descriptor(iface, i);
		if (ed == NULL) {
			printf("%s: could not read endpoint descriptor"
			       ": %s\n", devname, usbd_errstr(err));
			goto error;
		}

		addr = ed->bEndpointAddress;
		dir = UE_GET_DIR(ed->bEndpointAddress);
		attr = ed->bmAttributes & UE_XFERTYPE;
		if (dir == UE_DIR_IN && attr == UE_BULK)
			ucom->sc_bulkin_no = addr;
		else if (dir == UE_DIR_OUT && attr == UE_BULK)
			ucom->sc_bulkout_no = addr;
		else
			printf("%s: unexpected endpoint\n", devname);
	}
	if (ucom->sc_bulkin_no == -1) {
		printf("%s: Could not find data bulk in\n",
		       USBDEVNAME(ucom->sc_dev));
		goto error;
	}
	if (ucom->sc_bulkout_no == -1) {
		printf("%s: Could not find data bulk out\n",
		       USBDEVNAME(ucom->sc_dev));
		goto error;
	}

	TASK_INIT(&sc->sc_task, 0, ugensa_notify, sc);
	ucom_attach(&sc->sc_ucom);
	return 0;

error:
	DPRINTF(("ugensa_attach: ATTACH ERROR\n"));
	ucom->sc_dying = 1;
	return ENXIO;
}

static int
ugensa_detach(device_t self)
{
	struct ugensa_softc *sc = device_get_softc(self);
	int rv = 0;

	DPRINTF(("ugensa_detach: sc=%p\n", sc));

	sc->sc_ucom.sc_dying = 1;
	rv = ucom_detach(&sc->sc_ucom);

	return (rv);
}

static void
ugensa_notify(void *arg, int count)
{
	struct ugensa_softc *sc;

	sc = (struct ugensa_softc *)arg;
	if (sc->sc_ucom.sc_dying)
		return;
	ucom_status_change(&sc->sc_ucom);
}
